from django.conf import settings
from django.core.management.base import BaseCommand

from pucas.ldap import LDAPSearch, LDAPSearchException


class Command(BaseCommand):
    help = 'Look up one or more users in LDAP by netid'

    def add_arguments(self, parser):
        parser.add_argument('netid', nargs='+')
        parser.add_argument('--all', '-a', action='store_true',
            help='Retrieve all available LDAP attributes')

    def handle(self, *args, **options):
        ldap_search = LDAPSearch()
        for netid in options['netid']:
            self.stdout.write('\nLooking for %s...' % netid)
            try:
                info = ldap_search.find_user(netid, all_attributes=options['all'])
                # if all attributes were requested, just print the returned
                # ldap search object
                if options['all']:
                    self.stdout.write(info)
                # otherwise, display attributes configured in settings
                else:
                    for attr in settings.PUCAS_LDAP['ATTRIBUTES']:
                        self.stdout.write('%-15s %s' % (attr, getattr(info, attr)))
            except LDAPSearchException as err:
                self.stderr.write(self.style.ERROR(str(err)))




__version_info__ = (0, 9, 1, None)

# Dot-connect all but the last. Last is dash-connected if not None.
__version__ = ".".join([str(i) for i in __version_info__[:-1]])
if __version_info__[-1] is not None:
    __version__ += "-%s" % (__version_info__[-1],)



from django.dispatch import receiver
from django_cas_ng.signals import cas_user_authenticated

from pucas.ldap import user_info_from_ldap


@receiver(cas_user_authenticated)
def cas_login(sender, user, created, **kwargs):
    # Do an LDAP lookup to populate user data. Since CAS authentication
    # succeeded, should be no ambiguity in finding user info.

    # only populate attributes when a new user is created
    if created:
        user_info_from_ldap(user)





import ldap3
import logging
import importlib

from ldap3.core.exceptions import LDAPException, LDAPCursorError
from django.conf import settings


logger = logging.getLogger(__name__)


class LDAPSearchException(LDAPException):
    pass


class LDAPSearch(object):

    def __init__(self):
        # retrieve settings and initialize connection
        ldap_servers = []
        for server in settings.PUCAS_LDAP['SERVERS']:
            ldap_servers.append(ldap3.Server(server, get_info=ldap3.ALL,
                use_ssl=True))
        server_pool = ldap3.ServerPool(ldap_servers,
            ldap3.ROUND_ROBIN, active=True, exhaust=5)

        # Load username (in DN format) and password, if in settings
        bind_dn = settings.PUCAS_LDAP.get('BIND_DN', None)
        bind_password = settings.PUCAS_LDAP.get('BIND_PASSWORD', None)

        extra_args = {}
        # Use DN and password if set. Otherwise, use anononymous bind.
        if bind_dn and bind_password:
            extra_args.update({'user': bind_dn, 'password': bind_password})

        try:
            self.conn = ldap3.Connection(server_pool, auto_bind=True,
                                         **extra_args)
        except LDAPException as err:
            logging.error('Error establishing LDAP connection: %s', err)
            # re-raise to be caught elsewhere
            raise

    def find_user(self, netid, all_attributes=False):
        if netid:
            # check for required settings and error if not available
            required_configs = ['ATTRIBUTES', 'SEARCH_BASE', 'SEARCH_FILTER']
            if any(req not in settings.PUCAS_LDAP for req in required_configs):
                raise LDAPSearchException('LDAP is not configured for user lookup')

            # for testing, to see all available attributes
            if all_attributes:
                search_attributes = '*'
            else:
                search_attributes = settings.PUCAS_LDAP['ATTRIBUTES']

            self.conn.search(settings.PUCAS_LDAP['SEARCH_BASE'],
                    settings.PUCAS_LDAP['SEARCH_FILTER'] % {'user': netid},
                    attributes=search_attributes)
            if self.conn.entries:
                if len(self.conn.entries) > 1:
                    raise LDAPSearchException('Found more than one entry for %s' % netid)

                return self.conn.entries[0]

            else:
                raise LDAPSearchException('No match found for %s' % netid)

        else:
            raise LDAPSearchException('Error: requested LDAP lookup on empty netid')


def user_info_from_ldap(user):
    '''Populate django user info from ldap'''

    # configured mapping of user fields to ldap fields
    attr_map = settings.PUCAS_LDAP.get('ATTRIBUTE_MAP', None)
    # if no map is configured, nothing to do
    if not attr_map:
        # is logging sufficient here? or should it be an exception
        logging.warning('No attribute map configured; not populating user info'
                        ' from ldap')
        return

    user_info = LDAPSearch().find_user(user.username)
    if user_info:
        for user_attr, ldap_attr in attr_map.items():
            # Handle issues where an attribute may need to be populated by
            # multiple attributes OR where it is missing.

            # if just a string, convert to a list so handling can be uniform,
            # in cases where a field is multivalued.
            if isinstance(ldap_attr, str):
                ldap_attr = [ldap_attr]

            # iterate through the list items and break on the first one to
            # correct set without raising LDAPCursorError
            # This is a simplification for multivalued fields, since
            # typically we're mapping to only one value.
            for val in ldap_attr:
                try:
                    setattr(user, user_attr, str(getattr(user_info, val)))
                    break
                except LDAPCursorError:
                    pass
            # user getattr to check for a still unset value, in which case
            # set it to an empty string, used in situations where
            # a user is being updated -- and ot make sure values are
            # strings, not lists
            if not getattr(user, user_attr, None):
                setattr(user, user_attr, '')

        # optional custom user-init method set in django config
        extra_init = settings.PUCAS_LDAP.get('EXTRA_USER_INIT', None)
        if extra_init:
            # use importlib to import the module and get hte function
            mod_name, func_name = extra_init.rsplit('.', 1)
            mod = importlib.import_module(mod_name)
            extra_init_func = getattr(mod, func_name)
            extra_init_func(user, user_info)

        user.save()



from django.urls import path
from django_cas_ng.views import LoginView, LogoutView, CallbackView

urlpatterns = [
    path("login/", LoginView.as_view(), name="cas_ng_login"),
    path("logout/", LogoutView.as_view(), name="cas_ng_logout"),
    path("callback/", CallbackView.as_view(), name="cas_ng_proxy_callback"),
]



from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from pucas.ldap import LDAPSearch, LDAPSearchException, \
    user_info_from_ldap


class Command(BaseCommand):
    help = 'Initialize a new CAS user account'

    def add_arguments(self, parser):
        parser.add_argument('netids', nargs='+')
        parser.add_argument(
            '--admin',
            help='Give the specified user(s) superuser permissions (equivalent to createsuperuser)',
            action='store_true',
            default=False
        )
        parser.add_argument(
            '--staff',
            help='Give the specified user(s) staff permissions',
            action='store_true',
            default=False
        )

    def handle(self, *args, **options):
        User = get_user_model()

        ldap_search = LDAPSearch()
        netids = options['netids']
        admin = options['admin']
        staff = options['staff']
        for netid in netids:
            try:
                # make sure we can find the netid in LDAP first
                ldap_search.find_user(netid)
                user, created = User.objects.get_or_create(username=netid)
                # NOTE: should we re-init data from ldap even if user
                # already exists, or error?
                user_info_from_ldap(user)

                # If admin flag is set, make the user an admin
                if admin or staff:
                    user.is_staff = True
                    if admin:
                        user.is_superuser = True
                    user.save()

                self.stdout.write(
                    self.style.SUCCESS(
                        "%s user '%s'"
                        % ('Created' if created else 'Updated', netid)))

            except LDAPSearchException:
                self.stderr.write(
                    self.style.ERROR("LDAP information for '%s' not found"
                                     % netid))



import os
import sys
from setuptools import find_packages, setup
from pucas import __version__

with open(os.path.join(os.path.dirname(__file__), "README.rst")) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

REQUIREMENTS = ["django>=1.8", "django-cas-ng>=3.6", "ldap3"]
TEST_REQUIREMENTS = ["pytest", "pytest-django", "pytest-cov"]
if sys.version_info < (3, 0):
    TEST_REQUIREMENTS.append("mock")


setup(
    name="pucas",
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    license="Apache License, Version 2.0",
    description="Django app to login with CAS and populate user accounts with LDAP.",
    long_description=README,
    url="https://github.com/Princeton-CDH/django-pucas",
    install_requires=REQUIREMENTS,
    setup_requires=["pytest-runner"],
    tests_require=TEST_REQUIREMENTS,
    extras_require={
        "test": TEST_REQUIREMENTS,
    },
    author="CDH @ Princeton",
    author_email="digitalhumanities@princeton.edu",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: System :: Systems Administration :: Authentication/Directory",
        "Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP",
    ],
)



from io import StringIO
from unittest import mock

from django.conf import settings
from django.core.management import call_command
from django.test import TestCase, override_settings
from ldap3.core.exceptions import LDAPCursorError, LDAPException
import pytest

from pucas.ldap import LDAPSearch, LDAPSearchException, \
    user_info_from_ldap
from pucas.management.commands import createcasuser, ldapsearch
from pucas.signals import cas_login


class MockLDAPInfo(object):
    '''Simulate ldap result object with ldap3 specific behavior for getattr'''
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getattr__(self, attr):
        # __getattr__ only gets called when the default attribute access
        # falls through, so in this case, we always want that to raise the
        # cursor error
        raise LDAPCursorError


class TestMockLDAPInfo(TestCase):

    def test_init(self):
        mock_info = MockLDAPInfo(a=1, b=2)
        assert mock_info.a == 1
        assert mock_info.b == 2

    def test_getattr(self):
        mock_info = MockLDAPInfo(a=1, b=2)
        with pytest.raises(LDAPCursorError):
            foo = mock_info.c


class TestSignals(TestCase):
    # NOTE: using django TestCase for compatibility with
    # django test runner

    @mock.patch('pucas.signals.user_info_from_ldap')
    def test_cas_login(self, mock_userinit):
        mockuser = mock.Mock()
        # if create is not true, user init method should be called
        cas_login(mock.Mock(), mockuser, False)
        mock_userinit.assert_not_called()

        # if create is true, user init method should be called
        cas_login(mock.Mock(), mockuser, True)
        mock_userinit.assert_called_with(mockuser)


class TestLDAPSearch(TestCase):

    ldap_servers = ['lds81', 'ldap42', 'ld4all']
    dn = 'uid=foo,o= bar org,c=us'
    password = 'baz'

    @mock.patch('pucas.ldap.ldap3')
    @override_settings(PUCAS_LDAP={'SERVERS': ldap_servers})
    def test_init(self, mockldap3):
        # initialize and then check expected behavior against
        # mock ldap3
        LDAPSearch()

        test_servers = []
        for test_server in self.ldap_servers:
            mockldap3.Server.assert_any_call(test_server,
                get_info=mockldap3.ALL, use_ssl=True)

        # initialized servers are collected into server pool
        servers = [mockldap3.Server.return_value
                   for test_server in self.ldap_servers]
        mockldap3.ServerPool.assert_called_with(servers,
            mockldap3.ROUND_ROBIN, active=True, exhaust=5)

        # server pool is used for connection
        mockldap3.Connection.assert_called_with(mockldap3.ServerPool.return_value,
            auto_bind=True)

        with override_settings(PUCAS_LDAP={
            'SERVERS': self.ldap_servers,
            'BIND_DN': self.dn,
            'BIND_PASSWORD': self.password,
         }):
            LDAPSearch()
            # server pool is used for connection, now with password
            mockldap3.Connection.assert_called_with(mockldap3.ServerPool.return_value,
                auto_bind=True, user=self.dn, password=self.password)

        with pytest.raises(LDAPException):
            mockldap3.Connection.side_effect = LDAPException
            LDAPSearch()


    @mock.patch('pucas.ldap.ldap3')
    @override_settings(PUCAS_LDAP={'SERVERS': ldap_servers,
        'ATTRIBUTES': ['uid', 'sn', 'ou'],
        'SEARCH_BASE': 'o=my_org', 'SEARCH_FILTER': "(uid=%(user)s)"})
    def test_find_user(self, mockldap3):
        ldsearch = LDAPSearch()

        # empty netid should error
        with pytest.raises(LDAPSearchException):
            ldsearch.find_user(None)
        with pytest.raises(LDAPSearchException):
            ldsearch.find_user('')

        netid = 'jschmoe'
        # simulate no results
        ldsearch.conn.entries = []

        with pytest.raises(LDAPSearchException) as search_err:
            ldsearch.find_user(netid)
        assert 'No match found for %s' % netid in str(search_err.value)
        # search should use configured values
        ldsearch.conn.search.assert_called_with(settings.PUCAS_LDAP['SEARCH_BASE'],
            settings.PUCAS_LDAP['SEARCH_FILTER'] % {'user': netid},
            attributes=settings.PUCAS_LDAP['ATTRIBUTES'])

        # simulate too many matches
        ldsearch.conn.entries = [mock.Mock(), mock.Mock()]
        with pytest.raises(LDAPSearchException) as search_err:
            ldsearch.find_user(netid)

        assert 'Found more than one entry for %s' % netid in \
            str(search_err.value)

        # simulate one match
        userinfo = mock.Mock()
        ldsearch.conn.entries = [userinfo]
        assert ldsearch.find_user(netid) == userinfo

        # search for all attributes
        ldsearch.find_user(netid, all_attributes=True)
        # should use '*' instead of configured attributes
        ldsearch.conn.search.assert_called_with(settings.PUCAS_LDAP['SEARCH_BASE'],
            settings.PUCAS_LDAP['SEARCH_FILTER'] % {'user': netid},
            attributes='*')

        # with missing configs in any combination
        bad_settings = [
            # nothing set
            {},
            # attributes only
            {'ATTRIBUTES': ['foo']},
            # search filter missing
            {'ATTRIBUTES': ['foo'], 'SEARCH_BASE': 'u=foo'},
            # search base missing
            {'ATTRIBUTES': ['foo'], 'SEARCH_FILTER': '(uid=u)'},
        ]

        for bad_cfg in bad_settings:

            with override_settings(PUCAS_LDAP=bad_cfg):
                with pytest.raises(LDAPSearchException) as search_err:
                    ldsearch.find_user(netid)
            assert 'LDAP is not configured for user lookup' in str(search_err.value)


def extra_user_init(user, user_info):
    user.extra = 'custom init'


@mock.patch('pucas.ldap.LDAPSearch')
class TestUserInfo(TestCase):

    test_attr_map = {'first_name': 'givenName', 'last_name': 'surname',
                     'email': ['mail', 'eduPerson']}

    @override_settings(PUCAS_LDAP={})
    def test_no_attrs(self, mock_ldapsearch):
        mockuser = mock.Mock()

        user_info_from_ldap(mockuser)
        mock_ldapsearch.assert_not_called()

    @override_settings(PUCAS_LDAP={'ATTRIBUTE_MAP': test_attr_map})
    def test_attrs(self, mock_ldapsearch):
        mockuser = mock.Mock(username='jdoe')
        # simulate no user info returned
        mock_ldapsearch.return_value.find_user.return_value = None
        user_info_from_ldap(mockuser)
        # ldap search init should be called with no args
        mock_ldapsearch.assert_called_with()
        # find user should be called with username
        mock_ldapsearch.return_value.find_user.assert_called_with('jdoe')
        # user save should not be called - no data
        mockuser.save.assert_not_called()

        mock_ldapinfo = MockLDAPInfo(
                        eduPerson='jdoe2@example.com',
                        givenName='John',
                        surname='Doe',
                        mail='jdoe@example.com',
                        extra='foo'
                    )
        # first test that list style attributes are set in order, and string
        # attributes are set as given
        mock_ldapsearch.return_value.find_user.return_value = mock_ldapinfo
        user_info_from_ldap(mockuser)
        assert mockuser.first_name == mock_ldapinfo.givenName
        assert mockuser.last_name == mock_ldapinfo.surname
        assert mockuser.email == mock_ldapinfo.mail
        mockuser.save.assert_called_with()

        # second test that should pass over an unset eduPerson attr and
        # set using givenName in list
        # NOTE: recreating mock to clear all assigned attrs
        delattr(mock_ldapinfo, 'mail')
        user_info_from_ldap(mockuser)
        assert mockuser.first_name == mock_ldapinfo.givenName
        assert mockuser.last_name == mock_ldapinfo.surname
        assert mockuser.email == mock_ldapinfo.eduPerson
        mockuser.save.assert_called_with()

        # missing attribute altogether should result in an empty string
        delattr(mock_ldapinfo, 'givenName')
        delattr(mock_ldapinfo, 'surname')
        mockuser = mock.Mock(username='jdoe')
        # set to none to avoid Mock returning a mock and default behavior of
        # getattr
        mockuser.first_name = None
        mockuser.last_name = None
        mock_ldapsearch.return_value.find_user.return_value = mock_ldapinfo
        user_info_from_ldap(mockuser)
        assert mockuser.first_name == ''
        assert mockuser.last_name == ''
        assert mockuser.email == mock_ldapinfo.eduPerson
        mockuser.save.assert_called_with()

    @override_settings(PUCAS_LDAP={'ATTRIBUTE_MAP': test_attr_map,
        'EXTRA_USER_INIT': 'pucas.tests.extra_user_init'})
    def test_extra_init(self, mock_ldapsearch):
        mockuser = mock.Mock(username='jdoe')

        mock_ldapsearch.return_value.find_user.return_value = mock.Mock()
        user_info_from_ldap(mockuser)
        # check for custom field set by test extra init method
        assert mockuser.extra == 'custom init'
        mockuser.save.assert_called_with()


@mock.patch('pucas.management.commands.ldapsearch.LDAPSearch')
@override_settings(PUCAS_LDAP={'ATTRIBUTES': ['foo', 'bar', 'baz']})
class TestLDAPSearchCommand(TestCase):

    def setUp(self):
        self.cmd = ldapsearch.Command()
        # replace stdout/stderr with buffers to inspect output
        self.cmd.stdout = StringIO()
        self.cmd.stderr = StringIO()

    def test_search(self, mock_ldapsearch):
        mock_ldapinfo = mock.Mock(foo='phooey', bar='none', baz='1')
        mock_ldapsearch.return_value.find_user.return_value = mock_ldapinfo
        self.cmd.handle(netid=['jdoe'], all=False)
        mock_ldapsearch.assert_called_with()
        mock_ldapsearch.return_value.find_user.assert_called_with('jdoe',
            all_attributes=False)
        output = self.cmd.stdout.getvalue()
        assert 'Looking for jdoe...' in output
        assert '%-15s %s' % ('foo', 'phooey') in output
        assert '%-15s %s' % ('bar', 'none') in output
        assert '%-15s %s' % ('baz', '1') in output

    def test_search_all_attr(self, mock_ldapsearch):
        mock_ldapsearch.return_value.find_user.return_value = 'full return'
        self.cmd.handle(netid=['jdoe'], all=True)
        mock_ldapsearch.return_value.find_user.assert_called_with('jdoe',
            all_attributes=True)
        output = self.cmd.stdout.getvalue()
        # currently all attributes just prints the returned object
        assert 'full return' in output

    def test_err(self, mock_ldapsearch):
        error_message = 'Error looking for jdoe'
        mock_ldapsearch.return_value.find_user.side_effect = \
            LDAPSearchException(error_message)
        self.cmd.handle(netid=['jdoe'], all=False)
        output = self.cmd.stderr.getvalue()
        assert error_message in output

    def test_call_command(self, mock_ldapsearch):
        call_command('ldapsearch', 'jdoe')
        mock_ldapsearch.return_value.find_user.assert_called_with('jdoe',
            all_attributes=False)


@mock.patch('pucas.management.commands.createcasuser.get_user_model')
@mock.patch('pucas.management.commands.createcasuser.LDAPSearch')
@mock.patch('pucas.management.commands.createcasuser.user_info_from_ldap')
class TestCreateCasUserCommand(TestCase):

    def setUp(self):
        self.cmd = createcasuser.Command()
        self.cmd.stdout = StringIO()
        self.cmd.stderr = StringIO()

    def test_handle(self, mock_userinfo, mock_ldapsearch, mock_getuser):
        mockuser = mock.Mock(is_staff=False, is_superuser=False)
        mock_getuser.return_value.objects.get_or_create.return_value = \
            (mockuser, False)
        self.cmd.handle(netids=['jdoe'], admin=False, staff=False)
        # search should be called
        mock_ldapsearch.return_value.find_user.assert_called_with('jdoe')
        # user info method should be called
        mock_userinfo.assert_called_with(mockuser)
        # not given staff or superuser permissions
        assert not mockuser.is_staff
        assert not mockuser.is_superuser
        output = self.cmd.stdout.getvalue()
        assert "Updated user 'jdoe'" in output

        # init with staff=True
        self.cmd.handle(netids=['jdoe'], admin=False, staff=True)
        assert mockuser.is_staff
        # init with admin=True
        self.cmd.handle(netids=['jdoe'], admin=True, staff=True)
        assert mockuser.is_staff
        assert mockuser.is_admin

        # created vs updated
        mock_getuser.return_value.objects.get_or_create.return_value = \
            (mockuser, True)
        self.cmd.handle(netids=['jschmoe'], admin=True, staff=True)
        output = self.cmd.stdout.getvalue()
        assert "Created user 'jschmoe'" in output

    def test_err(self, mock_userinfo, mock_ldapsearch, mock_getuser):
        mock_ldapsearch.return_value.find_user.side_effect = LDAPSearchException
        self.cmd.handle(netids=['jdoe'], admin=False, staff=False)
        output = self.cmd.stderr.getvalue()
        assert "LDAP information for 'jdoe' not found" in output

    def test_call_command(self, mock_userinfo, mock_ldapsearch, mock_getuser):
        mock_ldapsearch.return_value.find_user.side_effect = LDAPSearchException
        call_command('createcasuser', 'jdoe', '--staff')
        mock_ldapsearch.return_value.find_user.assert_called_with('jdoe')
